"""Utility classes and functions."""
import re


def namespace_mapping(identifier, namespaces):
    """Get the namespace-mapped version of the ``identifier``.

    If the ``identifier`` has a namespace, then this is looked up in the ``namespaces`` dictionary. If the namespace
    is found, then it returns text in the format ``ns_localName``.

    :param identifier: The identifier to map
    :type identifier: ``string``
    :param namespaces: Namespaces to use for mapping
    :type namespaces: ``dict``
    """
    if namespaces and identifier:
        match = re.fullmatch(r'(?:\{([^}]+)\})?(.+)', identifier)
        if match:
            if match.group(1) in namespaces:
                if namespaces[match.group(1)]:
                    return f'{namespaces[match.group(1)]}_{match.group(2)}'
                else:
                    return match.group(2)
    return identifier



"""This is the main package for the Polymatheia library."""
try:
    from importlib.metadata import version
    __version__ = version('polymatheia')
except Exception:
    import pkg_resources
    __version__ = pkg_resources.get_distribution('polymatheia').version



r"""The :mod:`~polymatheia.transform` module implements a number of transform classes.

All transform classes use the following transform expression language and are designed to work with
:class:`~polymatheia.data.NavigableDict`. Each transform is structured as follows:
``('operator', 'parameter'[, 'parameter', ...])``. The number of parameters depends on the operator:

* Static: ``('static', target, value)``: The static ``value`` is stored at the location specified by the dotted
  path ``target``.
* Copy: ``('copy', target, source)``: The value specified by the dotted path ``source`` is stored at the location
  specified by the dotted path ``target``.
* Fill: ``('fill', target, value)``: If the value specified by the dotted path ``target`` is ``None``, then this is
  replaced by the ``value``, otherwise it is left untouched.
* Split: ``('split', target, splitter, source): The value specified by the dotted path ``source`` is split and
  stored at the location specified by the dotted path ``target``. If the value is a ``list``, then the list is split
  into its elements. If the value is a ``str``, then it is split using ``splitter`. Because this will produce more
  than one value in the transformation result, the ``target`` can contain the string ``{}``, which is replaced with
  the index of each split value (starting at 1).
* Combine: ``('combine', target, source1, ...)``: The values specified by the dotted paths ``source1`` (and optionally
  ``source2``, ``source3``,...) are combined into a list and stored at the location specified by the dotted path
  ``target``.
* Join: ``('join', target, joiner, source1, ...)``: The values specified by the dotted paths ``source1`` (and
  optionally ``source2``, ``source3``,...) are joined by the ``joiner`` and the result stored at the location specified
  by the dotted path ``target``. If only one ``source`` is specified, this is treated as a ``list`` and the values of
  the ``list`` joined using ``joiner``. If multiple ``source``\ s are specified, then the values returned by each
  dotted ``source`` path are returned.
* sequence: ``('sequence', expression1, expression2, ...)``: Applies the transform expressions in sequence. When using
  the sequence transformation, this should be the only top-level transformation being applied, otherwise the results
  cannot be guaranteed.
* parallel: ``('multiple', expression1, expression2, ...)``: Applies the transform expressions in parallel. Only needed
  if you want to convert multiple expressions in parallel inside a ``sequence`` transform. Parallel transforms **must**
  not be mixed with other transforms at the same level.
* custom: ``('custom', target, callable)``: Applies the custom transformation defined by the ``callable`` and stores
  the result in ``target``.
"""
from polymatheia.data import NavigableDict


class Transform(object):
    """The :class:`~polymatheia.transform.Transform` is a callable transformation for :class:`~polymatheia.data.NavigableDict`."""  # noqa: E501

    def __init__(self, mapping):
        """Create a new :class:`~polymatheia.transform.Transform`.

        :param mapping: The mappings that are applied when the :class:`~polymatheia.transform.Transform` is called. If a
                        ``list`` is passed as the parameter, then this constructs an implicit ``'parallel'`` transform.
        :type mapping: ``tuple`` or ``list``
        """
        if isinstance(mapping, list):
            mapping = ('parallel', *mapping)
        if mapping[0] == 'sequence':
            self._mapping = ('sequence', *[Transform(m) for m in mapping[1:]])
        elif mapping[0] == 'parallel':
            self._mapping = ('parallel', *[Transform(m) for m in mapping[1:]])
        else:
            self._mapping = mapping

    def __call__(self, record):
        """Transform the ``record`` according to the mappings of this :class:`~polymatheia.transform.Transform`.

        :param record: The record to transform
        :type record: :class:`~polymatheia.data.NavigableDict`
        :return: The transformed record
        :rtype: :class:`~polymatheia.data.NavigableDict`
        """
        result = NavigableDict({})
        if self._mapping[0] == 'copy':
            result.set(self._mapping[1], record.get(self._mapping[2]))
        elif self._mapping[0] == 'static':
            result.set(self._mapping[1], self._mapping[2])
        elif self._mapping[0] == 'fill':
            if record.get(self._mapping[1]) is None:
                result.set(self._mapping[1], self._mapping[2])
            else:
                result.set(self._mapping[1], record.get(self._mapping[1]))
        elif self._mapping[0] == 'split':
            value = record.get(self._mapping[3])
            if value:
                if isinstance(value, str):
                    for idx, part in enumerate(value.split(self._mapping[2])):
                        result.set(self._mapping[1].format(idx + 1), part)
                elif isinstance(value, list):
                    for idx, part in enumerate(value):
                        result.set(self._mapping[1].format(idx + 1), part)
        elif self._mapping[0] == 'combine':
            result.set(self._mapping[1], [record.get(path) for path in self._mapping[2:]])
        elif self._mapping[0] == 'join':
            if len(self._mapping) == 4:
                value = record.get(self._mapping[3])
                if value:
                    result.set(self._mapping[1], self._mapping[2].join(value))
            else:
                result.set(self._mapping[1], self._mapping[2].join([record.get(path) for path in self._mapping[3:]]))
        elif self._mapping[0] == 'sequence':
            tmp = record
            for part in self._mapping[1:]:
                result = part(tmp)
                tmp = result
        elif self._mapping[0] == 'parallel':
            for part in self._mapping[1:]:
                result.merge(part(record))
        elif self._mapping[0] == 'custom':
            result.set(self._mapping[1], self._mapping[2](record))
        return result


class RecordsTransform(object):
    """The :class:`~polymatheia.transform.RecordsTransform` provides a record transformation iterator container.

    The actual transformation is performed using a :class:`~polymatheia.transform.Transform` that is applied to
    each record in the source iterator.
    """

    def __init__(self, records, mappings):
        """Create a new :class:`~polymatheia.transform.RecordsTransform`.

        :param records: The source records to iterate over and transform
        :param mappings: The mappings that are applied to each record
        :type mappings: ``list``
        """
        self._records = records
        self._mappings = mappings

    def __iter__(self):
        """Return this class:`~polymatheia.transform.RecordsTransform` as the iterator."""
        return RecordsTransformIterator(self._records, self._mappings)


class RecordsTransformIterator(object):
    """The :class:`~polymatheia.transform.RecordsTransformIterator` provides a record transformation iterator.

    The actual transformation is performed using a :class:`~polymatheia.transform.Transform` that is applied to
    each record in the source iterator.
    """

    def __init__(self, records, mappings):
        """Create a new :class:`~polymatheia.transform.RecordsTransformIterator`.

        :param records: The source records to iterate over and transform
        :param mappings: The mappings that are applied to each record
        :type mappings: ``list``
        """
        self._it = iter(records)
        self._transform = Transform(mappings)

    def __iter__(self):
        """Return this class:`~polymatheia.transform.RecordsTransformIterator` as the iterator."""
        return self

    def __next__(self):
        """Transform and return the next :class:`~polymatheia.data.NavigableDict` record.

        :return: The transformed record
        :rtype: :class:`~polymatheia.data.NavigableDict`
        :raises StopIteration: If no more records are available
        """
        return self._transform(next(self._it))



"""The :mod:`~polymatheia.filter` module implements a number of filtering classes.

All filters support the same expression language and are designed to work with the
:class:`~polymatheia.data.NavigableDict`. Each individual expression is structured as follows:
``('operator', 'parameter'[, 'parameter', ...])``. The number of parameters depends on the operator:

* Equal: ``('eq', param1, param2)``: returns ``True`` if the two parameters are equal
* Not-equal: ``('neq', param1, param2)``: returns ``True`` if the two parameters are not equal
* Greater than: ``('gt', param1, param2)``: returns ``True`` if the param1 is greater than param2
* Greater than or equal: ``('gte', param1, param2)``: returns ``True`` if the param1 is greater than or equal to param2
* Less than: ``('lt', param1, param2)``: returns ``True`` if the param1 is less than param2
* Less than or equal: ``('lte', param1, param2)``: returns ``True`` if the param1 is less than or equal to param2
* Containment: ``('contains', param1, param2)``: returns ``True`` if the param2 is contained within param1
* Exists: ``('exists', param1)``: returns ``True`` if the param1 exists and has a value that is not ``None``
* Not: ``('not', expression)``: See below
* And: ``('and', expression1, expression2, ...)``: See below
* Or: ``('and', expression1, expression2, ...)``: See below
* True: ``('true',)``: always returns ``True``
* False: ``('true',)``: always returns ``False``

In the first six expressions both ``param1`` and ``param2`` can be one of the following three, allowing for comparison
of a value from a :class:`~polymatheia.data.NavigableDict` record against a constant value, two values from the same
:class:`~polymatheia.data.NavigableDict` record, or two constant values:

* A constant value: The comparison is performed against this value
* A dotted path: The comparison is performed against the value fetched from the record using the path
* A a list: The comparison is performed against the value fetched from the record using the path in list form. This
  is needed if either the path elements contain a full-stop or square brackets or if you wish to compare against a
  value
  at the top level of the :class:`~polymatheia.data.NavigableDict` record.

For the operators ``'not'``, ``'and'``, and ``'or'`` the ``expression`` parameter is a nested expression using any of
the given operators. The ``'not'`` operator only supports at most a single nested expression. The ``'and'`` and
``'or'`` operators allow any number of nested expression parameters. This

* ``'not'``, ``'and'``, or ``'or'`` with no nested expressions always return ``True``
* ``'and'`` or ``'or'`` with a single nested expression always return the value of the nested expression
* ``'and'`` or ``'or'`` with more than one nested expression are evaluated lazily. That means that for ``'and'`` as
  soon as a nested expression is ``False``, ``False`` is immediately returned. Likewise for ``'or'`` as soon as
  a nested expression is ``True``, then ``True`` is immediately returned.
"""


class Filter(object):
    """The :class:`~polymatheia.filter.Filter` is a callable filter for :class:`~polymatheia.data.NavigableDict`."""

    def __init__(self, expression):
        """Create a new :class:`~polymatheia.filter.Filter`.

        :param expression: The expression that this :class:`~polymatheia.filter.Filter` will apply.
        :type expression: ``tuple``
        """
        if expression[0] == 'not':
            if len(expression) == 1:
                self._expression = ('true',)
            else:
                self._expression = ('not', Filter(expression[1]))
        elif expression[0] == 'and':
            if len(expression) == 2:
                self._expression = expression[1]
            else:
                self._expression = tuple(['and'] + [Filter(exp) for exp in expression[1:]])
        elif expression[0] == 'or':
            if len(expression) == 2:
                self._expression = expression[1]
            else:
                self._expression = tuple(['or'] + [Filter(exp) for exp in expression[1:]])
        else:
            self._expression = expression

    def __call__(self, record):
        """Apply the expression to the ``record``.

        :param record: The record to apply the filter expression to.
        :type record: :class:`~polymatheia.data.NavigableDict`
        :return: Whether the filter expression matches the record or not.
        :rtype: ``bool``
        """
        if self._expression[0] == 'eq':
            return self._get_value(record, self._expression[1]) == self._get_value(record, self._expression[2])
        elif self._expression[0] == 'neq':
            return self._get_value(record, self._expression[1]) != self._get_value(record, self._expression[2])
        elif self._expression[0] == 'gt':
            return self._get_value(record, self._expression[1]) > self._get_value(record, self._expression[2])
        elif self._expression[0] == 'gte':
            return self._get_value(record, self._expression[1]) >= self._get_value(record, self._expression[2])
        elif self._expression[0] == 'lt':
            return self._get_value(record, self._expression[1]) < self._get_value(record, self._expression[2])
        elif self._expression[0] == 'lte':
            return self._get_value(record, self._expression[1]) <= self._get_value(record, self._expression[2])
        elif self._expression[0] == 'contains':
            try:
                return self._get_value(record, self._expression[2]) in self._get_value(record, self._expression[1])
            except TypeError:
                return False
        elif self._expression[0] == 'exists':
            return self._get_value(record, self._expression[1]) is not None
        elif self._expression[0] == 'not':
            return not self._expression[1](record)
        elif self._expression[0] == 'and':
            for part in self._expression[1:]:
                if not part(record):
                    return False
            return True
        elif self._expression[0] == 'or':
            all_false = False
            for part in self._expression[1:]:
                if part(record):
                    return True
                else:
                    all_false = True
            return not all_false
        elif self._expression[0] == 'true':
            return True
        elif self._expression[0] == 'false':
            return False
        return False

    def _get_value(self, record, path):
        if isinstance(path, str) and ('.' in path or ('[' in path and ']' in path)):
            return record.get(path)
        elif isinstance(path, list):
            return record.get(path)
        else:
            return path


class RecordsFilter(object):
    """The :class:`~polymatheia.filter.RecordsFilter` provides a records filter iteration container.

    It allows iterating over all :class:`~polymatheia.data.NavigableDict` that match the given filter expression.
    Filtering is performed using a :class:`~polymatheia.filter.Filter`.
    """

    def __init__(self, records, expression):
        """Create a new :class:`~polymatheia.filter.RecordsFilter`.

        :param records: The source records to iterate over and filter.
        :param expression: The filter expression to apply
        :type expression: ``tuple`` or :class:`~polymatheia.filter.Filter`
        """
        self._records = records
        self._expression = expression

    def __iter__(self):
        """Return a new class:`~polymatheia.filter.RecordsFilterIterator` as the iterator."""
        return RecordsFilterIterator(self._records, self._expression)


class RecordsFilterIterator(object):
    """The :class:`~polymatheia.filter.RecordsFilterIterator` provides a records filter iterator.

    It allows iterating over all :class:`~polymatheia.data.NavigableDict` that match the given filter expression.
    Filtering is performed using a :class:`~polymatheia.filter.Filter`.
    """

    def __init__(self, records, expression):
        """Create a new :class:`~polymatheia.filter.RecordsFilterIterator`.

        :param records: The source records to iterate over and filter.
        :param expression: The filter expression to apply
        :type expression: ``tuple`` or :class:`~polymatheia.filter.Filter`
        """
        self._it = iter(records)
        if isinstance(expression, Filter):
            self._expression = expression
        else:
            self._expression = Filter(expression)

    def __iter__(self):
        """Return this class:`~polymatheia.filter.RecordsFilterIterator` as the iterator."""
        return self

    def __next__(self):
        """Return the next :class:`~polymatheia.data.NavigableDict` that matches the filter.

        :return: The next record that matches the filter
        :rtype: :class:`~polymatheia.data.NavigableDict`
        :raises StopIteration: If no more records are available
        """
        record = next(self._it)
        while not self._expression(record):
            record = next(self._it)
        return record



"""The classes in the :mod:`polymatheia.data` package handle data input and output."""
import json

from polymatheia.util import namespace_mapping


class NavigableDict(dict):
    """The :class:`~polymatheia.data.NavigableDict` is a ``dict`` subclass that allows access via dot notation.

    >>> test = NavigableDict(one='1')
    >>> test.one
    1
    >>> test['one']
    1

    The :class:`~polymatheia.data.NavigableDict` works like a ``dict`` in any other respect.
    """

    def __init__(self, *args, **kwargs):
        """Initialise the :class:`~polymatheia.data.NavigableDict`, ensuring that data is coerced."""
        self.update(*args, **kwargs)

    def __getattr__(self, key):
        """Retrieve the value with the given ``key``.

        :return: The value for the ``key``
        :raises KeyError: If no value exists for ``key``
        """
        return self[key]

    def __setattr__(self, key, value):
        """Set the ``value`` for ``key``.

        This method automatically coerces any ``dict`` ``value`` into :class:`~polymatheia.data.NavigableDict`.

        :param key: The key to set the value for
        :type key: ``string``
        :param value: The value to set
        """
        if isinstance(value, dict):
            if isinstance(value, NavigableDict):
                self[key] = value
            else:
                self[key] = NavigableDict(value)
        elif isinstance(value, list):
            self[key] = [v if isinstance(v, NavigableDict) else
                         NavigableDict(v) if isinstance(v, dict)
                         else v for v in value]
        else:
            self[key] = value

    def __delattr__(self, key):
        """Delete the value with the given ``key``.

        :raises KeyError: If no value exists for ``key``
        """
        del self[key]

    def __str__(self):
        """Return a pretty-printed JSON representation."""
        return json.dumps(self, indent=2)

    def update(self, *args, **kwargs):
        """Update the content of this :class:`~polymatheia.data.NavigableDict`.

        Any ``dict`` passed will be coerced into :class:`~polymatheia.data.NavigableDict`
        """
        for key, value in dict(*args, **kwargs).items():
            setattr(self, key, value)

    def get(self, path, default=None):
        r"""Get the value specified by the ``path``.

        The ``path`` can either be a ``str``, in which case it is split into its component parts. If it is a ``list``
        then it is used as is. The following ``str`` ``path`` structures are supported:

        * ``x``: Get the value with the key ``'x'``
        * ``x.y``: Get the value with the key ``'x'`` and then within that the key ``'y'``
        * ``x.a``: Get the list with the key ``'x'`` and in that returns the ``a``\ -th element in the list
        * ``x[a]``: Get the list with the key ``'x'`` and in that returns the ``a``\ -th element in the list

        In general the ``list`` format is only needed if one of the parts used in the ``path`` contains a ``'.'``,
        ``'['``, or ``']'``.

        This differs from just using attribute access, in how it handles lists. If a value is a list and the next path
        element is not a list index, then it will return a list, applying the remainder of the path to each element
        in the list.

        :param path: The path to get the value for
        :type path: ``str`` or ``[str, ...]``
        :param default: The default value to return if the ``path`` does not identify a value
        :return: The value identified by ``path`` or the ``default`` value
        """
        if isinstance(path, str):
            path = self._split_path(path)
        if len(path) == 1:
            if path[0] in self:
                return self[path[0]]
        elif len(path) > 1:
            if path[0] in self:
                child = self[path[0]]
                if isinstance(child, NavigableDict):
                    return child.get(path[1:], default=default)
                elif isinstance(child, list):
                    try:
                        element = child[int(path[1])]
                        if len(path) == 2:
                            return element
                        elif isinstance(element, NavigableDict):
                            return element.get(path[2:], default=default)
                    except ValueError:
                        result = []
                        for element in child:
                            if isinstance(element, NavigableDict):
                                result.append(element.get(path[1:]))
                            else:
                                result.append(default)
                        return result
                    except IndexError:
                        pass
        return default

    def set(self, path, value):
        r"""Set the ``value`` at the location specified by ``path``.

        The ``path`` can either be a ``str``, in which case it is split into its component parts. If it is a ``list``
        then it is used as is. The following ``str`` ``path`` structures are supported:

        * ``x``: Set the value with the key ``'x'``
        * ``x.y``: Set the value with the key ``'x'`` and then within that the key ``'y'``
        * ``x.a``: Set the list with the key ``'x'`` and in that set the ``a``\ -th element in the list
        * ``x[a]``: Set the list with the key ``'x'`` and in that set the ``a``\ -th element in the list

        In general the ``list`` format is only needed if one of the parts used in the ``path`` contains a ``'.'``,
        ``'['``, or ``']'``.

        :param path: The path to set the value for
        :type path: ``str`` or ``[str, ...]``
        :param value: The value to set
        """
        if isinstance(path, str):
            path = self._split_path(path)
        tmp = self
        for idx, element in enumerate(path):
            if idx == len(path) - 1:
                if isinstance(tmp, list):
                    tmp[int(element)] = value
                else:
                    setattr(tmp, element, value)
            else:
                if isinstance(tmp, list):
                    tmp = tmp[int(element)]
                else:
                    if element not in tmp:
                        tmp[element] = NavigableDict({})
                    tmp = tmp[element]

    def merge(self, other):
        """Merge the ``other`` values.

        Unlike the :meth:`~polymatheia.data.NavigableDict.update` method, this method does not immediately overwrite
        any existing values. Instead if both the new and existing value are :class:`~polymatheia.data.NavigableDict`,
        then the values from the ``other`` :class:`~polymatheia.data.NavigableDict` are merged into the existing
        :class:`~polymatheia.data.NavigableDict`. Similarly, if both the new and existing value are ``list``, then
        the new existing list is extended with the new list. If neither of these conditions hold, then the existing
        value is overwritten by the new value.
        """
        if isinstance(other, dict) and not isinstance(other, NavigableDict):
            other = NavigableDict(other)
        for key, value in other.items():
            if key in self:
                if isinstance(self[key], NavigableDict) and isinstance(value, NavigableDict):
                    self[key].merge(value)
                elif isinstance(self[key], list) and isinstance(value, list):
                    self[key].extend(value)
                else:
                    setattr(self, key, value)
            else:
                setattr(self, key, value)

    def _split_path(self, path):
        """Split the ``path`` into a ``tuple``.

        The ``path`` is split on ``'.'``. Additionally, if list indexes are identified via ``[...]``, then the
        ``path`` is also split on those.

        :param path: The path to split into its parts
        :type path: ``str``
        :return: The ``path`` as a ``tuple``
        :rtype: ``tuple`` of ``str``
        """
        result = []
        tmp = []
        for character in path:
            if character == '.':
                if tmp:
                    result.append(''.join(tmp))
                tmp = []
            elif character == '[':
                if tmp:
                    result.append(''.join(tmp))
                tmp = []
            elif character == ']':
                if tmp:
                    result.append(''.join(tmp))
                tmp = []
            else:
                tmp.append(character)
        if tmp:
            result.append(''.join(tmp))
        return tuple(result)


class NavigableDictIterator(object):
    """The :class:`~polymatheia.data.NavigableDictIterator` maps values to :class:`~polymatheia.data.NavigableDict`.

    If the iterator it wraps returns ``dict`` objects, then these are simply converted into
    :class:`~polymatheia.data.NavigableDict` objects. If a ``mapper`` function is provided, then this function is
    called with each value and it must return a ``dict`` object that is then converted into a
    :class:`~polymatheia.data.NavigableDict` object. Otherwise a :class:`~polymatheia.data.NavigableDict` is returned
    that has a single key ``value``, with the wrapped iterator value.
    """

    def __init__(self, it, mapper=None):
        """Create a new :class:`~polymatheia.data.NavigableDictIterator`.

        :param it: The iterator that provides the values.
        :param mapper: An optional mapping function converting a single iterator value into a ``dict`` object.
        """
        self._it = it
        self._mapper = mapper

    def __iter__(self):
        """Return this :class:`~polymatheia.data.NavigableDictIterator` as the iterator."""
        return self

    def __next__(self):
        """Return the next :class:`~polymatheia.data.NavigableDict`.

        :raises StopIteration: If no more :class:`~polymatheia.data.NavigableDict` are available
        """
        value = next(self._it)
        if self._mapper:
            return NavigableDict(self._mapper(value))
        elif isinstance(value, dict):
            return NavigableDict(value)
        else:
            return NavigableDict({'value': value})


class LimitingIterator(object):
    """The :class:`~polymatheia.data.LimitingIterator` limits the number of records returned from a wrapped iterator."""

    def __init__(self, it, max_records):
        """Create a new :class:`~polymatheia.data.LimitingIterator`.

        :param it: The wrapped iterator.
        :param max_records: The maximum number of records to return.
        :type max_records: ``number``
        """
        self._it = it
        self._max_records = max_records

    def __iter__(self):
        """Return this :class:`~polymatheia.data.LimitingIterator` as the iterator."""
        return self

    def __next__(self):
        """Return the next value.

        :raises StopIteration: If no more values are available
        """
        if self._max_records <= 0:
            raise StopIteration()
        self._max_records = self._max_records - 1
        return next(self._it)


def xml_to_navigable_dict(node):
    r"""Convert the XML node to a dictionary.

    Child ``node``\ s are returned as keys of the dictionary. Where a ``node`` occurs multiple times, the key returns
    a list of dictionaries for the children. This means that the ordering information in the original XML document is
    **lost**. All attributes are available via the ``_attrib`` key. The ``node``\ 's text is available via the
    ``_text`` key and any text that directly follows the ``node`` is available via the ``_tail`` key.

    :param node: The XML node to convert
    :return: The dictionary representation of the XML node
    :rtype: :class:`~polymatheia.data.NavigableDict`
    """
    namespaces = dict([(v, k) for k, v in node.nsmap.items()])
    tmp = {}
    if node.text:
        tmp['_text'] = node.text
    if node.tail:
        tmp['_tail'] = node.tail
    if len(node.attrib) > 0:
        tmp['_attrib'] = {}
    for key, value in node.attrib.items():
        tmp['_attrib'][namespace_mapping(key, namespaces)] = value
    for child in node:
        if namespace_mapping(child.tag, namespaces) in tmp:
            if not isinstance(tmp[namespace_mapping(child.tag, namespaces)], list):
                tmp[namespace_mapping(child.tag, namespaces)] = [tmp[namespace_mapping(child.tag, namespaces)]]
            tmp[namespace_mapping(child.tag, namespaces)].append(xml_to_navigable_dict(child))
        else:
            tmp[namespace_mapping(child.tag, namespaces)] = xml_to_navigable_dict(child)
    return NavigableDict(tmp)



"""This module provides a range of readers for accessing local and remote resources.

All readers return their data as :class:`~polymatheia.data.NavigableDict`.
"""
import json
import os

from csv import DictReader
from lxml import etree
from requests import get
from sickle import Sickle
from srupy import SRUpy

from polymatheia.data import NavigableDict, NavigableDictIterator, LimitingIterator, xml_to_navigable_dict


class OAIMetadataFormatReader(object):
    """The class:`~polymatheia.data.reader.OAIMetadataFormatReader` is a container for OAI-PMH MetadataFormat.

    The underlying library automatically handles the continuation parameters, allowing for simple iteration.
    """

    def __init__(self, url):
        """Construct a new class:`~polymatheia.data.reader.OAIMetadataFormatReader`.

        :param url: The base URL of the OAI-PMH server
        :type url: ``str``
        """
        self._url = url

    def __iter__(self):
        """Return a new class:`~polymatheia.data.NavigableDictIterator` as the iterator."""
        return NavigableDictIterator(Sickle(self._url).ListMetadataFormats(),
                                     mapper=lambda meta_format: {'schema': meta_format.schema,
                                                                 'metadataPrefix': meta_format.metadataPrefix,
                                                                 'metadataNamespace': meta_format.metadataNamespace})


class OAISetReader(object):
    """The class:`~polymatheia.data.reader.OAISetReader` is an iteration container for OAI-PMH Sets.

    The underlying library automatically handles the continuation parameters, allowing for simple iteration.
    """

    def __init__(self, url):
        """Construct a new class:`~polymatheia.data.reader.OAISetReader`.

        :param url: The base URL of the OAI-PMH server
        :type url: ``str``
        """
        self._url = url
        self._it = Sickle(url).ListSets()

    def __iter__(self):
        """Return a new class:`~polymatheia.data.NavigableDictIterator` as the iterator."""
        return NavigableDictIterator(Sickle(self._url).ListSets(),
                                     mapper=lambda oai_set: {'setSpec': oai_set.setSpec, 'setName': oai_set.setName})


class OAIRecordReader(object):
    """The :class:`~polymatheia.data.reader.OAIRecordReader` is an iteration container for OAI-PMH Records.

    The underlying library automatically handles the continuation parameters, allowing for simple iteration.
    """

    def __init__(self, url, metadata_prefix='oai_dc', max_records=None, set_spec=None):
        """Construct a new :class:`~polymatheia.data.reader.OAIRecordReader`.

        :param url: The base URL of the OAI-PMH server
        :type url: ``str``
        :param metadataPrefix: The metadata prefix to use for accessing data
        :type metadataPrefix: ``str``
        :param max_records: The maximum number of records to return. Default (``None``) returns all records
        :type max_records: ``int``
        :param set_spec: The OAI Set specification for limiting which metadata to fetch
        :type set_spec: ``str``
        """
        self._url = url
        self._metadata_prefix = metadata_prefix
        self._set_spec = set_spec
        self._max_records = max_records

    def __iter__(self):
        """Return a new class:`~polymatheia.data.NavigableDictIterator` as the iterator.

        If ``max_records`` is set, then the class:`~polymatheia.data.NavigableDictIterator` is wrapped in a
        class:`~polymatheia.data.LimitingIterator`.
        """
        it = NavigableDictIterator(Sickle(self._url).ListRecords(metadataPrefix=self._metadata_prefix,
                                                                 set=self._set_spec,
                                                                 ignore_deleted=True),
                                   mapper=lambda record: xml_to_navigable_dict(etree.fromstring(
                                       record.raw,
                                       parser=etree.XMLParser(remove_comments=True))))
        if self._max_records is not None:
            it = LimitingIterator(it, self._max_records)
        return it


class JSONReader():
    """The :class:`~polymatheia.data.reader.JSONReader` is a container for reading JSON files from the filesystem.

    It is designed to provide access to data serialised using the :class:`~polymatheia.data.writer.JSONWriter`.

    .. important::

       It does **not** guarantee that the order of records is the same as the order in which they were written
       to the local filesystem.
    """

    def __init__(self, directory):
        """Create a new :class:`~polymatheia.data.reader.JSONReader`.

        :param directory: The base directory within which to load the files
        :type directory: ``str``
        """
        self._directory = directory
        self._filelist = []
        for basepath, _, filenames in os.walk(directory):
            for filename in filenames:
                if filename.endswith('.json'):
                    self._filelist.append(os.path.join(basepath, filename))

    def __iter__(self):
        """Return a new :class:`~polymatheia.data.NavigableDictIterator` as the iterator."""
        return NavigableDictIterator(iter(self._filelist),
                                     mapper=self._load)

    def _load(self, filename):
        """Return the next file as a :class:`~polymatheia.data.NavigableDict`."""
        with open(filename) as in_f:
            return json.load(in_f)


class XMLReader():
    """The :class:`~polymatheia.data.reader.XMLReader` is a container for reading XML files from the local filesystem.

    The :class:`~polymatheia.data.reader.XMLReader` will only load files that have a ".xml" extension.
    """

    def __init__(self, directory):
        """Create a new :class:`~polymatheia.data.reader.XMLReader`.

        :param directory: The base directory within which to load the files
        :type directory: ``str``
        """
        self._directory = directory
        self._filelist = []
        for basepath, _, filenames in os.walk(directory):
            for filename in filenames:
                if filename.endswith('.xml'):
                    self._filelist.append(os.path.join(basepath, filename))

    def __iter__(self):
        """Return a new :class:`~polymatheia.data.NavigableDictIterator` as the iterator."""
        return NavigableDictIterator(iter(self._filelist),
                                     mapper=self._load)

    def _load(self, filename):
        """Return the next file as a :class:`~polymatheia.data.NavigableDict`."""
        with open(filename) as in_f:
            return xml_to_navigable_dict(etree.parse(in_f, parser=etree.XMLParser(remove_comments=True)).getroot())


class EuropeanaSearchReader(object):
    """The :class:`~polymatheia.data.reader.EuropeanaSearchReader` provides access to the Europeana Search API.

    The initial search is run immediately on creating a new :class:`~polymatheia.data.reader.EuropeanaSearchReader`.
    The iterator will automatically paginate through the full set of result pages.

    .. attribute:: result_count
       :type: int

       The total number of records returned by the search.

    .. attribute:: facets
       :type: polymatheia.data.NavigableDict

       The facets generated by the search. This is only set if the ``profile`` parameter is set to ``'facets'``.
    """

    def __init__(self, api_key, query, max_records=None, query_facets=None, media=None, thumbnail=None,
                 reusability=None, profile=None):
        """Create a new :class:`~polymatheia.data.reader.EuropeanaSearchReader`.

        :param api_key: The Europeana API key
        :type api_key: ``str``
        :param query: The query string
        :type query: ``str``
        :param max_records: The maximum number of records to return. Defaults to all records
        :type max_records: ``int``
        :param query_facets: The list of query facets to apply to the search
        :type query_facets: ``list`` of ``str``
        :param media: Whether to require that matching records have media attached. Defaults to no requirement
        :type media: ``bool``
        :param thumbnail: Whether to require that matching records have a thumbnail. Defaults to no requirement
        :type thumbnail: ``bool``
        :param reusability: The reusability (rights) to require. Defaults to no limits
        :type reusability: ``str``
        :param profile: The result profile to request. Defaults to ``'standard'``
        :type profile: ``str``
        """
        self._api_key = api_key
        self._query = query
        self._max_records = max_records
        self._cursor = '*'
        self._offset = 0
        self._query_facets = query_facets
        self._media = media
        self._thumbnail = thumbnail
        self._reusability = reusability
        self._profile = profile
        it = iter(self)
        self.result_count = it.result_count
        self.facets = it.facets

    def __iter__(self):
        """Return this :class:`~polymatheia.data.reader.EuropeanaSearchReader` as the iterator."""
        return EuropeanaSearchIterator(self._api_key, self._query, self._max_records, self._query_facets, self._media,
                                       self._thumbnail, self._reusability, self._profile)


class EuropeanaSearchIterator(object):
    """The :class:`~polymatheia.data.reader.EuropeanaSearchIterator` provides an iterator for the Europeana Search API.

    The initial search is run immediately on creating a new :class:`~polymatheia.data.reader.EuropeanaSearchIterator`.
    The iterator will automatically paginate through the full set of result pages.

    .. attribute:: result_count
       :type: int

       The total number of records returned by the search.

    .. attribute:: facets
       :type: polymatheia.data.NavigableDict

       The facets generated by the search. This is only set if the ``profile`` parameter is set to ``'facets'``.
    """

    def __init__(self, api_key, query, max_records=None, query_facets=None, media=None, thumbnail=None,
                 reusability=None, profile=None):
        """Create a new :class:`~polymatheia.data.reader.EuropeanaSearchReader`.

        :param api_key: The Europeana API key
        :type api_key: ``str``
        :param query: The query string
        :type query: ``str``
        :param max_records: The maximum number of records to return. Defaults to all records
        :type max_records: ``int``
        :param query_facets: The list of query facets to apply to the search
        :type query_facets: ``list`` of ``str``
        :param media: Whether to require that matching records have media attached. Defaults to no requirement
        :type media: ``bool``
        :param thumbnail: Whether to require that matching records have a thumbnail. Defaults to no requirement
        :type thumbnail: ``bool``
        :param reusability: The reusability (rights) to require. Defaults to no limits
        :type reusability: ``str``
        :param profile: The result profile to request. Defaults to ``'standard'``
        :type profile: ``str``
        """
        self._api_key = api_key
        self._query = query
        self._max_records = max_records
        self._cursor = '*'
        self._offset = 0
        self._query_facets = query_facets
        self._media = media
        self._thumbnail = thumbnail
        self._reusability = reusability
        self._profile = profile
        self.result_count = 0
        self.facets = None
        self._run_search()

    def __iter__(self):
        """Return this :class:`~polymatheia.data.reader.EuropeanaSearchIterator` as the iterator."""
        return self

    def __next__(self):
        """Return the next record as a :class:`~polymatheia.data.NavigableDict`.

        :raises StopIteration: If no more Records are available
        """
        if self._max_records is not None:
            self._max_records = self._max_records - 1
            if self._max_records < 0:
                raise StopIteration()
        try:
            return NavigableDict(next(self._it))
        except StopIteration:
            if self._offset < self.result_count:
                self._run_search()
                return NavigableDict(next(self._it))
            else:
                raise StopIteration()

    def _run_search(self):
        """Run the actual search query."""
        params = [('wskey', self._api_key),
                  ('query', self._query),
                  ('cursor', self._cursor)]
        if self._max_records and self._max_records < 50:
            params.append(('rows', self._max_records))
        else:
            params.append(('rows', 50))
        if self._query_facets:
            params.extend([('qf', qf) for qf in self._query_facets])
        if self._media is not None:
            params.append(('media', 'true' if self._media else 'false'))
        if self._thumbnail is not None:
            params.append(('thumbnail', 'true' if self._thumbnail else 'false'))
        if self._reusability is not None:
            params.append(('reusability', self._reusability))
        if self._profile is not None:
            params.append(('profile', self._profile))
        response = get('https://api.europeana.eu/record/v2/search.json', params=params)
        if response.status_code == 200:
            data = response.json()
            self._it = iter(data['items'])
            self.result_count = data['totalResults']
            self._offset = self._offset + data['itemsCount']
            if 'nextCursor' in data:
                self._cursor = data['nextCursor']
            if 'facets' in data:
                self.facets = [NavigableDict(facet) for facet in data['facets']]
        else:
            raise Exception(response.json()['error'])


class CSVReader(object):
    """The :class:`~polymatheia.data.reader.CSVReader` provides access to a CSV file."""

    def __init__(self, source):
        """Create a new :class:`~polymatheia.data.reader.CSVReader`.

        :param source: The source to load the CSV from. Can either be a ``str`` filename or a file-like object
        """
        if isinstance(source, str):
            self._file = open(source)
        else:
            self._file = source

    def __iter__(self):
        """Return this :class:`~polymatheia.data.reader.CSVReader` as the iterator."""
        if self._file.seekable():
            self._file.seek(0)
        return NavigableDictIterator(iter(DictReader(self._file)))


class SRUExplainRecordReader(object):
    """The class:`~polymatheia.data.reader.SRUExplainRecordReader` is a container for SRU Explain Records."""

    def __init__(self, url):
        """Construct a new class:`~polymatheia.data.reader.SRUExplainRecordReader`.

        :param url: The base URL of the SRU server
        :type url: ``str``
        """
        self._url = url
        self._explain = SRUpy(self._url).explain()
        self.schemas = [(schema["@name"], schema.title)
                        for schema in NavigableDict(self._explain).explain.schemaInfo.schema]
        self.echo = NavigableDict(self._explain.echo)

    def __iter__(self):
        """Return a new class:`~polymatheia.data.NavigableDictIterator` as the iterator."""
        return NavigableDictIterator(iter(self._explain),
                                     mapper=lambda record: {record[0]: record[1]}
                                     )


class SRURecordReader(object):
    """The :class:`~polymatheia.data.reader.SRURecordReader` is an iteration container for Records fetched via SRU.

    The underlying library (SRUpy) automatically handles the continuation parameters, allowing for simple iteration.
    """

    def __init__(self, url, query, max_records=None, record_schema="dc", **kwargs):
        """Construct a new :class:`~polymatheia.data.reader.SRURecordReader`.

        :param url: The base URL of the SRU endpoint
        :type url: ``str``
        :param query: The query string
        :type query: ``str``
        :param max_records: The maximum number of records to return
        :type max_records: ``int``
        :param record_schema: Schema in which records will be returned. Defaults to Dublin Core schema.
        :type record_schema: ``str``
        :param kwargs: Additional request parameters that will be sent to the SRU server
        """
        self._url = url
        self._query = query
        self._max_records = max_records
        self._record_schema = record_schema
        self._kwargs = kwargs
        self.record_count = None
        self.echo = None

    def __iter__(self):
        """Return a new class:`~polymatheia.data.NavigableDictIterator` as the iterator."""
        sru_records = SRUpy(self._url).get_records(query=self._query,
                                                   maximumRecords=self._max_records,
                                                   recordSchema=self._record_schema,
                                                   **self._kwargs)
        self.record_count = sru_records.number_of_records
        if sru_records.echo:
            self.echo = NavigableDict(sru_records.echo)
        return NavigableDictIterator(sru_records,
                                     mapper=lambda record: xml_to_navigable_dict(
                                         etree.fromstring(
                                             record.raw,
                                             parser=etree.XMLParser(remove_comments=True)
                                         )
                                     ))

    @staticmethod
    def result_count(url, query):
        """Return result count for the given query.

        :param url: The base URL of the SRU endpoint
        :type url: ``str``
        :param query: The query string
        :type query: ``str``
        """
        return SRUpy(url).get_records(query=query, maximumRecords=1).number_of_records



"""This module provides a writer for serialising data to the local filesystem."""
import json
import os
import re

from csv import DictWriter
from hashlib import sha256
from lxml import etree
from pandas import DataFrame


class JSONWriter():
    """The :class:`~polymatheia.data.writer.JSONWriter` writes records to the local filesystem as JSON files."""

    def __init__(self, directory, id_path):
        """Create a new :class:`~polymatheia.data.writer.JSONWriter`.

        For each record the identifier is used to create a directory structure. In the leaf directory the identifier
        is then used as the filename.

        :param directory: The base directory within which to create the files
        :type directory: ``str``
        :param id_path: The path used to access the identifier in the record
        :type id_path: ``str`` or ``list``
        """
        self._directory = directory
        if isinstance(id_path, str):
            self._id_path = id_path.split('.')
        else:
            self._id_path = id_path

    def write(self, records):
        """Write the records to the file-system.

        :param records: The records to write
        :type records: Iterable of :class:`~polymatheia.data.NavigableDict`
        """
        for record in records:
            identifier = record.get(self._id_path)
            if identifier:
                hash = sha256(identifier.encode('utf-8'))
                hex = hash.hexdigest()
                file_path = os.path.join(
                    self._directory,
                    *[hex[idx:idx+4] for idx in range(0, len(hex), 4)],
                    hex)
                os.makedirs(os.path.dirname(file_path), exist_ok=True)
                with open(f'{file_path}.json', 'w') as out_f:
                    json.dump(record, out_f)


class XMLWriter():
    """The :class:`~polymatheia.data.writer.XMLWriter` writes records to the local filesystem as XML."""

    def __init__(self, directory, id_path):
        """Create a new :class:`~polymatheia.data.writer.XMLWriter`.

        For each record the identifier is used to create a directory structure. In the leaf directory the identifier
        is then used as the filename.

        :param directory: The base directory within which to create the files
        :type directory: ``str``
        :param id_path: The path used to access the identifier in the record
        :type id_path: ``str`` or ``list``
        """
        self._directory = directory
        if isinstance(id_path, str):
            self._id_path = id_path.split('.')
        else:
            self._id_path = id_path

    def write(self, records):
        """Write the records to the file-system.

        :param records: The records to write
        :type records: Iterable of :class:`~polymatheia.data.NavigableDict`
        """
        for record in records:
            identifier = record.get(self._id_path)
            if identifier:
                hash = sha256(identifier.encode('utf-8'))
                hex = hash.hexdigest()
                file_path = os.path.join(
                    self._directory,
                    *[hex[idx:idx+4] for idx in range(0, len(hex), 4)],
                    hex)
                os.makedirs(os.path.dirname(file_path), exist_ok=True)
                with open(f'{file_path}.xml', 'wb') as out_f:
                    root = etree.Element('record')
                    self._build_xml_doc(root, record)
                    out_f.write(etree.tostring(root))

    def _build_xml_doc(self, parent, data):
        """Build the XML document tree.

        Tag names are generated from the keys in the ``data``, ensuring that they are valid XML tag names.

        Handles nested ``data`` trees by nesting elements and lists by generating the same tag repeatedly.

        :param parent: The parent node to attach elements to
        :type parent: :class:`~lxml.etree.Element`
        :param data: The data to build the tree from
        :type data: :class:`~polymatheia.data.NavigableDict`
        """
        for key, value in data.items():
            if isinstance(value, list):
                for sub_value in value:
                    element = etree.Element(self._valid_xml_tag(key))
                    if isinstance(sub_value, dict):
                        self._build_xml_doc(element, sub_value)
                    else:
                        element.text = str(value)
                    parent.append(element)
            elif isinstance(value, dict):
                element = etree.Element(self._valid_xml_tag(key))
                self._build_xml_doc(element, value)
                parent.append(element)
            else:
                element = etree.Element(self._valid_xml_tag(key))
                element.text = str(value)
                parent.append(element)

    def _valid_xml_tag(self, tag):
        """Generate a valid XML tag for the given ``tag``.

        :param tag: The tag to generate a valid XML tag for
        :type tag: ``str``
        :return: A valid XML tag
        :rtype: ``str``
        """
        tag = re.sub(r'\s+', '-', tag)
        tag = ''.join(re.findall(r'\w|\d|-|_|\.', tag))
        while not re.match(r'\w', tag) or re.match(r'\d', tag):
            tag = tag[1:]
        if tag.lower().startswith('xml'):
            tag = tag[3:]
        return tag


class CSVWriter():
    """The :class:`~polymatheia.data.writer.CSVWriter` writes records into a CSV file.

    The :class:`~polymatheia.data.writer.CSVWriter` assumes that no record contains any kind of nested data. If
    it is passed nested data, then the behaviour is undefined.
    """

    def __init__(self, target, default_value='', extras_action='ignore', column_names=None):
        """Create a new :class:`~polymatheia.data.writer.CSVWriter`.

        :param target: The target to write the CSV to. Can either be a ``str`` filename or an existing file-like object
        :param default_value: The default value to output if a record does not contain a value for a specified CSV
                              column name
        :param extras_action: The action to take if a record contains keys that are not in the CVS fieldnames. Set to
                              ``'ignore'`` to just ignore this (the default). Set to ``'raise'`` to raise a
                              ``ValueError``.
        :type extras_action: ``str``
        :param fieldnames: The CSV column names to use. If ``None`` is specified, then the column names are derived
                           from the first record's keys.
        :type fieldnames: ``list`` of ``str``
        """
        self._target = target
        self._default_value = default_value
        self._extras_action = extras_action
        self._column_names = column_names

    def write(self, records):
        """Write the ``records`` to the CSV file.

        :param records: The records to write
        :type records: Iterable of :class:`~polymatheia.data.NavigableDict`
        """
        if isinstance(self._target, str):
            with open(self._target, 'w') as out_file:
                self._write_csv(out_file, records)
        else:
            self._write_csv(self._target, records)

    def _write_csv(self, file, records):
        """Perform the actual writing of the CSV file.

        :param file: The file-like object to write to
        :param records: The records to write
        :type records: Iterable of :class:`~polymatheia.data.NavigableDict`
        """
        csv_writer = None
        for record in records:
            if not csv_writer:
                csv_writer = DictWriter(file,
                                        fieldnames=record.keys() if self._column_names is None
                                        else self._column_names,
                                        restval=self._default_value,
                                        extrasaction=self._extras_action)
                csv_writer.writeheader()
            csv_writer.writerow(record)


class PandasDFWriter(object):
    """The :class:`~polymatheia.data.writer.PandasDFWriter` writes records to a Pandas :class:`~pandas.DataFrame`.

    The :class:`~polymatheia.data.writer.PandasDFWriter` attempts to automatically coerce columns to integers or
    floats.

    The :class:`~polymatheia.data.writer.PandasDFWriter` assumes that no record contains any kind of nested data. If
    it is passed nested data, then the behaviour is undefined.
    """

    def __init__(self):
        """Create a new :class:`~polymatheia.data.writer.PandasDFWriter`."""
        self.df = None

    def write(self, records):
        """Write the ``records`` to the Pandas :class:`~pandas.DataFrame`.

        :param records: The records to write
        :type records: Iterable of :class:`~polymatheia.data.NavigableDict`
        :return: The Pandas dataframe
        :rtype: :class:`~pandas.DataFrame`
        """
        columns = {}
        for record in records:
            for key, value in record.items():
                if key not in columns:
                    columns[key] = []
                columns[key].append(value)
        for key, values in columns.items():
            try:
                columns[key] = [int(v) for v in values]
            except ValueError:
                try:
                    columns[key] = [float(v) for v in values]
                except ValueError:
                    pass
        return DataFrame(columns)
